# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager
import logging
import os
from paramiko import SSHClient, SSHConfig, MissingHostKeyPolicy
from paramiko.ssh_exception import SSHException, NoValidConnectionsError
import shutil
import signal
import socket
import stat
import tempfile
import warnings

from ducktape.utils.http_utils import HttpMixin
from ducktape.utils.util import wait_until
from ducktape.errors import DucktapeError


def check_ssh(method):
    def wrapper(self, *args, **kwargs):
        try:
            return method(self, *args, **kwargs)
        except (SSHException, NoValidConnectionsError, socket.error) as e:
            if self._custom_ssh_exception_checks:
                self._log(logging.DEBUG, "caught ssh error", exc_info=True)
                self._log(logging.DEBUG, "starting ssh checks:")
                self._log(logging.DEBUG, "\n".join(repr(f) for f in self._custom_ssh_exception_checks))
                for func in self._custom_ssh_exception_checks:
                    func(e, self)
            raise
    return wrapper


class RemoteAccountSSHConfig(object):
    def __init__(self, host=None, hostname=None, user=None, port=None, password=None, identityfile=None,
                 connecttimeout=None, **kwargs):
        """Wrapper for ssh configs used by ducktape to connect to remote machines.

        The fields in this class are lowercase versions of a small selection of ssh config properties
        (see man page: "man ssh_config")
        """
        self.host = host
        self.hostname = hostname or 'localhost'
        self.user = user
        self.port = port or 22
        self.port = int(self.port)
        self.password = password
        self.identityfile = identityfile
        # None is default, and it means default TCP timeout will be used.
        self.connecttimeout = int(connecttimeout) if connecttimeout is not None else None

    @staticmethod
    def from_string(config_str):
        """Construct RemoteAccountSSHConfig object from a string that looks like

        Host the-host
            Hostname the-hostname
            Port 22
            User ubuntu
            IdentityFile /path/to/key
        """
        config = SSHConfig()
        config.parse(config_str.split("\n"))

        hostnames = config.get_hostnames()
        if '*' in hostnames:
            hostnames.remove('*')
        assert len(hostnames) == 1, "Expected hostnames to have single entry: %s" % hostnames
        host = hostnames.pop()

        config_dict = config.lookup(host)
        if config_dict.get("identityfile") is not None:
            # paramiko.SSHConfig parses this in as a list, but we only want a single string
            config_dict["identityfile"] = config_dict["identityfile"][0]

        return RemoteAccountSSHConfig(host, **config_dict)

    def to_json(self):
        return self.__dict__

    def __repr__(self):
        return str(self.to_json())

    def __eq__(self, other):
        return other and other.__dict__ == self.__dict__

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))


class RemoteAccountError(DucktapeError):
    """This exception is raised when an attempted action on a remote node fails.
    """

    def __init__(self, account, msg):
        self.account_str = str(account)
        self.msg = msg

    def __str__(self):
        return "%s: %s" % (self.account_str, self.msg)


class RemoteCommandError(RemoteAccountError):
    """This exception is raised when a process run by ssh*() returns a non-zero exit status.
    """

    def __init__(self, account, cmd, exit_status, msg):
        self.account_str = str(account)
        self.exit_status = exit_status
        self.cmd = cmd
        self.msg = msg

    def __str__(self):
        msg = "%s: Command '%s' returned non-zero exit status %d." % (self.account_str, self.cmd, self.exit_status)
        if self.msg:
            msg += " Remote error message: %s" % self.msg
        return msg


class RemoteAccount(HttpMixin):
    """RemoteAccount is the heart of interaction with cluster nodes,
    and every allocated cluster node has a reference to an instance of RemoteAccount.

    It wraps metadata such as ssh configs, and provides methods for file system manipulation and shell commands.

    Each operating system has its own RemoteAccount implementation.
    """

    def __init__(self, ssh_config, externally_routable_ip=None, logger=None, ssh_exception_checks=[]):
        # Instance of RemoteAccountSSHConfig - use this instead of a dict, because we need the entire object to
        # be hashable
        self.ssh_config = ssh_config

        # We don't want to rely on the hostname (e.g. 'worker1') having been added to the driver host's /etc/hosts file.
        # But that means we need to distinguish between the hostname and the value of hostname we use for SSH commands.
        # We try to satisfy all use cases and keep things simple by
        #   a) storing the hostname the user probably expects (the "Host" value in .ssh/config)
        #   b) saving the real value we use for running the SSH command
        self.hostname = ssh_config.host
        self.ssh_hostname = ssh_config.hostname

        self.user = ssh_config.user
        self.externally_routable_ip = externally_routable_ip
        self._logger = logger
        self.os = None
        self._ssh_client = None
        self._sftp_client = None
        self._custom_ssh_exception_checks = ssh_exception_checks

    @property
    def operating_system(self):
        return self.os

    @property
    def logger(self):
        if self._logger:
            return self._logger
        else:
            return logging.getLogger(__name__)

    @logger.setter
    def logger(self, logger):
        self._logger = logger

    def _log(self, level, msg, *args, **kwargs):
        msg = "%s: %s" % (str(self), msg)
        self.logger.log(level, msg, *args, **kwargs)

    @check_ssh
    def _set_ssh_client(self):
        client = SSHClient()
        client.set_missing_host_key_policy(IgnoreMissingHostKeyPolicy())

        self._log(logging.DEBUG, "ssh_config: %s" % str(self.ssh_config))

        client.connect(
            hostname=self.ssh_config.hostname,
            port=self.ssh_config.port,
            username=self.ssh_config.user,
            password=self.ssh_config.password,
            key_filename=self.ssh_config.identityfile,
            look_for_keys=False,
            timeout=self.ssh_config.connecttimeout)

        if self._ssh_client:
            self._ssh_client.close()
        self._ssh_client = client
        self._set_sftp_client()

    @property
    def ssh_client(self):
        if (self._ssh_client
                and self._ssh_client.get_transport()
                and self._ssh_client.get_transport().is_active()):
            try:
                transport = self._ssh_client.get_transport()
                transport.send_ignore()
            except Exception as e:
                self._log(logging.DEBUG, "exception getting ssh_client (creating new client): %s" % str(e))
                self._set_ssh_client()
        else:
            self._set_ssh_client()

        return self._ssh_client

    def _set_sftp_client(self):
        if self._sftp_client:
            self._sftp_client.close()
        self._sftp_client = self.ssh_client.open_sftp()

    @property
    def sftp_client(self):
        if not self._sftp_client:
            self._set_sftp_client()
        else:
            self.ssh_client  # test connection

        return self._sftp_client

    def close(self):
        """Close/release any outstanding network connections to remote account."""

        if self._ssh_client:
            self._ssh_client.close()
            self._ssh_client = None
        if self._sftp_client:
            self._sftp_client.close()
            self._sftp_client = None

    def __str__(self):
        r = ""
        if self.user:
            r += self.user + "@"
        r += self.hostname
        return r

    def __repr__(self):
        return str(self.__dict__)

    def __eq__(self, other):
        return other is not None and self.__dict__ == other.__dict__

    def __hash__(self):
        return hash(tuple(sorted(self.__dict__.items())))

    def wait_for_http_service(self, port, headers, timeout=20, path='/'):
        """Wait until this service node is available/awake."""
        url = "http://%s:%s%s" % (self.externally_routable_ip, str(port), path)

        err_msg = "Timed out trying to contact service on %s. " % url + \
            "Either the service failed to start, or there is a problem with the url."
        wait_until(lambda: self._can_ping_url(url, headers), timeout_sec=timeout, backoff_sec=.25, err_msg=err_msg)

    def _can_ping_url(self, url, headers):
        """See if we can successfully issue a GET request to the given url."""
        try:
            self.http_request(url, "GET", None, headers, timeout=.75)
            return True
        except Exception:
            return False

    def available(self):
        # TODO: https://github.com/confluentinc/ducktape/issues/339
        # try:
        #     self.ssh_client
        # except Exception:
        #     return False
        # else:
        #     return True
        # finally:
        #     self.close()
        return True

    @check_ssh
    def ssh(self, cmd, allow_fail=False):
        """Run the given command on the remote host, and block until the command has finished running.

        :param cmd: The remote ssh command
        :param allow_fail: If True, ignore nonzero exit status of the remote command,
               else raise an ``RemoteCommandError``

        :return: The exit status of the command.
        :raise RemoteCommandError: If allow_fail is False and the command returns a non-zero exit status
        """
        self._log(logging.DEBUG, "Running ssh command: %s" % cmd)

        client = self.ssh_client
        stdin, stdout, stderr = client.exec_command(cmd)

        # Unfortunately we need to read over the channel to ensure that recv_exit_status won't hang. See:
        # http://docs.paramiko.org/en/2.0/api/channel.html#paramiko.channel.Channel.recv_exit_status
        stdout.read()
        exit_status = stdout.channel.recv_exit_status()
        try:
            if exit_status != 0:
                if not allow_fail:
                    raise RemoteCommandError(self, cmd, exit_status, stderr.read())
                else:
                    self._log(logging.DEBUG, "Running ssh command '%s' exited with status %d and message: %s" %
                              (cmd, exit_status, stderr.read()))
        finally:
            stdin.close()
            stdout.close()
            stderr.close()

        return exit_status

    @check_ssh
    def ssh_capture(self, cmd, allow_fail=False, callback=None, combine_stderr=True, timeout_sec=None):
        """Run the given command asynchronously via ssh, and return an SSHOutputIter object.

        Does *not* block

        :param cmd: The remote ssh command
        :param allow_fail: If True, ignore nonzero exit status of the remote command,
               else raise an ``RemoteCommandError``
        :param callback: If set, the iterator returns ``callback(line)``
               for each line of output instead of the raw output
        :param combine_stderr: If True, return output from both stderr and stdout of the remote process.
        :param timeout_sec: Set timeout on blocking reads/writes. Default None. For more details see
            http://docs.paramiko.org/en/2.0/api/channel.html#paramiko.channel.Channel.settimeout

        :return SSHOutputIter: object which allows iteration through each line of output.
        :raise RemoteCommandError: If ``allow_fail`` is False and the command returns a non-zero exit status
        """
        self._log(logging.DEBUG, "Running ssh command: %s" % cmd)

        client = self.ssh_client
        chan = client.get_transport().open_session(timeout=timeout_sec)

        chan.settimeout(timeout_sec)
        chan.exec_command(cmd)
        chan.set_combine_stderr(combine_stderr)

        stdin = chan.makefile('wb', -1)  # set bufsize to -1
        stdout = chan.makefile('r', -1)
        stderr = chan.makefile_stderr('r', -1)

        def output_generator():

            for line in iter(stdout.readline, ''):

                if callback is None:
                    yield line
                else:
                    yield callback(line)
            try:
                exit_status = stdout.channel.recv_exit_status()
                if exit_status != 0:
                    if not allow_fail:
                        raise RemoteCommandError(self, cmd, exit_status, stderr.read())
                    else:
                        self._log(logging.DEBUG, "Running ssh command '%s' exited with status %d and message: %s" %
                                  (cmd, exit_status, stderr.read()))
            finally:
                stdin.close()
                stdout.close()
                stderr.close()

        return SSHOutputIter(output_generator, stdout)

    @check_ssh
    def ssh_output(self, cmd, allow_fail=False, combine_stderr=True, timeout_sec=None):
        """Runs the command via SSH and captures the output, returning it as a string.

        :param cmd: The remote ssh command.
        :param allow_fail: If True, ignore nonzero exit status of the remote command,
               else raise an ``RemoteCommandError``
        :param combine_stderr: If True, return output from both stderr and stdout of the remote process.
        :param timeout_sec: Set timeout on blocking reads/writes. Default None. For more details see
            http://docs.paramiko.org/en/2.0/api/channel.html#paramiko.channel.Channel.settimeout

        :return: The stdout output from the ssh command.
        :raise RemoteCommandError: If ``allow_fail`` is False and the command returns a non-zero exit status
        """
        self._log(logging.DEBUG, "Running ssh command: %s" % cmd)

        client = self.ssh_client
        chan = client.get_transport().open_session(timeout=timeout_sec)

        chan.settimeout(timeout_sec)
        chan.exec_command(cmd)
        chan.set_combine_stderr(combine_stderr)

        stdin = chan.makefile('wb', -1)  # set bufsize to -1
        stdout = chan.makefile('r', -1)
        stderr = chan.makefile_stderr('r', -1)

        try:
            stdoutdata = stdout.read()
            exit_status = stdin.channel.recv_exit_status()
            if exit_status != 0:
                if not allow_fail:
                    raise RemoteCommandError(self, cmd, exit_status, stderr.read())
                else:
                    self._log(logging.DEBUG, "Running ssh command '%s' exited with status %d and message: %s" %
                              (cmd, exit_status, stderr.read()))
        finally:
            stdin.close()
            stdout.close()
            stderr.close()

        return stdoutdata

    def alive(self, pid):
        """Return True if and only if process with given pid is alive."""
        try:
            self.ssh("kill -0 %s" % str(pid), allow_fail=False)
            return True
        except Exception:
            return False

    def signal(self, pid, sig, allow_fail=False):
        cmd = "kill -%d %s" % (int(sig), str(pid))
        self.ssh(cmd, allow_fail=allow_fail)

    def kill_process(self, process_grep_str, clean_shutdown=True, allow_fail=False):
        cmd = """ps ax | grep -i """ + process_grep_str + """ | grep -v grep | awk '{print $1}'"""
        pids = [pid for pid in self.ssh_capture(cmd, allow_fail=True)]

        if clean_shutdown:
            sig = signal.SIGTERM
        else:
            sig = signal.SIGKILL

        for pid in pids:
            self.signal(pid, sig, allow_fail=allow_fail)

    def java_pids(self, match):
        """
        Get all the Java process IDs matching 'match'.

        :param match:               The AWK expression to match
        """
        cmd = """jcmd | awk '/%s/ { print $1 }'""" % match
        return [int(pid) for pid in self.ssh_capture(cmd, allow_fail=True)]

    def kill_java_processes(self, match, clean_shutdown=True, allow_fail=False):
        """
        Kill all the java processes matching 'match'.

        :param match:               The AWK expression to match
        :param clean_shutdown:      True if we should shut down cleanly with SIGTERM;
                                    false if we should shut down with SIGKILL.
        :param allow_fail:          True if we should throw exceptions if the ssh commands fail.
        """
        cmd = """jcmd | awk '/%s/ { print $1 }'""" % match
        pids = [pid for pid in self.ssh_capture(cmd, allow_fail=True)]

        if clean_shutdown:
            sig = signal.SIGTERM
        else:
            sig = signal.SIGKILL

        for pid in pids:
            self.signal(pid, sig, allow_fail=allow_fail)

    def copy_between(self, src, dest, dest_node):
        """Copy src to dest on dest_node

        :param src: Path to the file or directory we want to copy
        :param dest: The destination path
        :param dest_node: The node to which we want to copy the file/directory

        Note that if src is a directory, this will automatically copy recursively.

        """
        # TODO: if dest is an existing file, what is the behavior?

        temp_dir = tempfile.mkdtemp()

        try:
            # TODO: deal with very unlikely case that src_name matches temp_dir name?
            # TODO: I think this actually works
            local_dest = self._re_anchor_basename(src, temp_dir)

            self.copy_from(src, local_dest)

            dest_node.account.copy_to(local_dest, dest)

        finally:
            if os.path.isdir(temp_dir):
                shutil.rmtree(temp_dir)

    def scp_from(self, src, dest, recursive=False):
        warnings.warn("scp_from is now deprecated. Please use copy_from")
        self.copy_from(src, dest)

    def _re_anchor_basename(self, path, directory):
        """Anchor the basename of path onto the given directory

        Helper for the various copy_* methods.

        :param path: Path to a file or directory. Could be on the driver machine or a worker machine.
        :param directory: Path to a directory. Could be on the driver machine or a worker machine.

        Example::

            path/to/the_basename, another/path/ -> another/path/the_basename
        """
        path_basename = path

        # trim off path separator from end of path
        # this is necessary because os.path.basename of a path ending in a separator is an empty string
        # For example:
        #   os.path.basename("the/path/") == ""
        #   os.path.basename("the/path") == "path"
        if path_basename.endswith(os.path.sep):
            path_basename = path_basename[:-len(os.path.sep)]
        path_basename = os.path.basename(path_basename)

        return os.path.join(directory, path_basename)

    @check_ssh
    def copy_from(self, src, dest):
        if os.path.isdir(dest):
            # dest is an existing directory, so assuming src looks like path/to/src_name,
            # in this case we'll copy as:
            #   path/to/src_name -> dest/src_name
            dest = self._re_anchor_basename(src, dest)

        if self.isfile(src):
            self.sftp_client.get(src, dest)
        elif self.isdir(src):
            # we can now assume dest path looks like: path_that_exists/new_directory
            os.mkdir(dest)

            # for obj in `ls src`, if it's a file, copy with copy_file_from, elif its a directory, call again
            for obj in self.sftp_client.listdir(src):
                obj_path = os.path.join(src, obj)
                if self.isfile(obj_path) or self.isdir(obj_path):
                    self.copy_from(obj_path, dest)
                else:
                    # TODO what about uncopyable file types?
                    pass

    def scp_to(self, src, dest, recursive=False):
        warnings.warn("scp_to is now deprecated. Please use copy_to")
        self.copy_to(src, dest)

    @check_ssh
    def copy_to(self, src, dest):

        if self.isdir(dest):
            # dest is an existing directory, so assuming src looks like path/to/src_name,
            # in this case we'll copy as:
            #   path/to/src_name -> dest/src_name
            dest = self._re_anchor_basename(src, dest)

        if os.path.isfile(src):
            # local to remote
            self.sftp_client.put(src, dest)
        elif os.path.isdir(src):
            # we can now assume dest path looks like: path_that_exists/new_directory
            self.mkdir(dest)

            # for obj in `ls src`, if it's a file, copy with copy_file_from, elif its a directory, call again
            for obj in os.listdir(src):
                obj_path = os.path.join(src, obj)
                if os.path.isfile(obj_path) or os.path.isdir(obj_path):
                    self.copy_to(obj_path, dest)
                else:
                    # TODO what about uncopyable file types?
                    pass

    @check_ssh
    def islink(self, path):
        try:
            # stat should follow symlinks
            path_stat = self.sftp_client.lstat(path)
            return stat.S_ISLNK(path_stat.st_mode)
        except Exception:
            return False

    @check_ssh
    def isdir(self, path):
        try:
            # stat should follow symlinks
            path_stat = self.sftp_client.stat(path)
            return stat.S_ISDIR(path_stat.st_mode)
        except Exception:
            return False

    @check_ssh
    def exists(self, path):
        """Test that the path exists, but don't follow symlinks."""
        try:
            # stat follows symlinks and tries to stat the actual file
            self.sftp_client.lstat(path)
            return True
        except IOError:
            return False

    @check_ssh
    def isfile(self, path):
        """Imitates semantics of os.path.isfile

        :param path: Path to the thing to check
        :return: True if path is a file or a symlink to a file, else False. Note False can mean path does not exist.
        """
        try:
            # stat should follow symlinks
            path_stat = self.sftp_client.stat(path)
            return stat.S_ISREG(path_stat.st_mode)
        except Exception:
            return False

    def open(self, path, mode='r'):
        return self.sftp_client.open(path, mode)

    @check_ssh
    def create_file(self, path, contents):
        """Create file at path, with the given contents.

        If the path already exists, it will be overwritten.
        """
        # TODO: what should semantics be if path exists? what actually happens if it already exists?
        # TODO: what happens if the base part of the path does not exist?

        with self.sftp_client.open(path, "w") as f:
            f.write(contents)

    _DEFAULT_PERMISSIONS = int('755', 8)

    @check_ssh
    def mkdir(self, path, mode=_DEFAULT_PERMISSIONS):
        self.sftp_client.mkdir(path, mode)

    def mkdirs(self, path, mode=_DEFAULT_PERMISSIONS):
        self.ssh("mkdir -p %s && chmod %o %s" % (path, mode, path))

    def remove(self, path, allow_fail=False):
        """Remove the given file or directory"""

        if allow_fail:
            cmd = "rm -rf %s" % path
        else:
            cmd = "rm -r %s" % path

        self.ssh(cmd, allow_fail=allow_fail)

    @contextmanager
    def monitor_log(self, log):
        """
        Context manager that returns an object that helps you wait for events to
        occur in a log. This checks the size of the log at the beginning of the
        block and makes a helper object available with convenience methods for
        checking or waiting for a pattern to appear in the log. This will commonly
        be used to start a process, then wait for a log message indicating the
        process is in a ready state.

        See ``LogMonitor`` for more usage information.
        """
        try:
            offset = int(self.ssh_output("wc -c %s" % log).split()[0])
        except Exception:
            offset = 0
        yield LogMonitor(self, log, offset)


class SSHOutputIter(object):
    """Helper class that wraps around an iterable object to provide has_next() in addition to next()
    """

    def __init__(self, iter_obj_func, channel_file=None):
        """
        :param iter_obj_func: A generator that returns an iterator over stdout from the remote process
        :param channel_file: A paramiko ``ChannelFile`` object
        """
        self.iter_obj_func = iter_obj_func
        self.iter_obj = iter_obj_func()
        self.channel_file = channel_file

        # sentinel is used as an indicator that there is currently nothing cached
        # If self.cached is self.sentinel, then next object from ier_obj is not yet cached.
        self.sentinel = object()
        self.cached = self.sentinel

    def __iter__(self):
        return self

    def next(self):
        if self.cached is self.sentinel:
            return next(self.iter_obj)
        next_obj = self.cached
        self.cached = self.sentinel
        return next_obj

    __next__ = next

    def has_next(self, timeout_sec=None):
        """Return True if next(iter_obj) would return another object within timeout_sec, else False.

        If timeout_sec is None, next(iter_obj) may block indefinitely.
        """
        assert timeout_sec is None or self.channel_file is not None, "should have descriptor to enforce timeout"

        prev_timeout = None
        if self.cached is self.sentinel:
            if self.channel_file is not None:
                prev_timeout = self.channel_file.channel.gettimeout()

                # when timeout_sec is None, next(iter_obj) will block indefinitely
                self.channel_file.channel.settimeout(timeout_sec)
            try:
                self.cached = next(self.iter_obj, self.sentinel)
            except socket.timeout:
                self.iter_obj = self.iter_obj_func()
                self.cached = self.sentinel
            finally:
                if self.channel_file is not None:
                    # restore preexisting timeout
                    self.channel_file.channel.settimeout(prev_timeout)

        return self.cached is not self.sentinel


class LogMonitor(object):
    """
    Helper class returned by monitor_log. Should be used as::

        with remote_account.monitor_log("/path/to/log") as monitor:
            remote_account.ssh("/command/to/start")
            monitor.wait_until("pattern.*to.*grep.*for", timeout_sec=5)

    to run the command and then wait for the pattern to appear in the log.
    """

    def __init__(self, acct, log, offset):
        self.acct = acct
        self.log = log
        self.offset = offset

    def wait_until(self, pattern, **kwargs):
        """
        Wait until the specified pattern is found in the log, after the initial
        offset recorded when the LogMonitor was created. Additional keyword args
        are passed directly to ``ducktape.utils.util.wait_until``
        """
        return wait_until(lambda: self.acct.ssh("tail -c +%d %s | grep '%s'" % (self.offset + 1, self.log, pattern),
                                                allow_fail=True) == 0, **kwargs)


class IgnoreMissingHostKeyPolicy(MissingHostKeyPolicy):
    """Policy for ignoring missing host keys.
    Many examples show use of AutoAddPolicy, but this clutters up the known_hosts file unnecessarily.
    """

    def missing_host_key(self, client, hostname, key):
        return



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
from ducktape.cluster.cluster_spec import ClusterSpec
from ducktape.tests.test import Test
from ducktape.mark import ignore, parametrize
from ducktape.mark.resource import cluster


_flake = False


class TestThingy(Test):
    """Fake ducktape test class"""

    def min_cluster_spec(self):
        """ This test uses many nodes, wow!"""
        return ClusterSpec.simple_linux(1000)

    def test_pi(self):
        return {"data": 3.14159}

    def test_delayed(self):
        time.sleep(1)

    @ignore
    def test_ignore1(self):
        pass

    @ignore(x=5)
    @parametrize(x=5)
    def test_ignore2(self, x=2):
        pass

    def test_failure(self):
        raise Exception("This failed")

    def test_flaky(self):
        global _flake
        flake, _flake = _flake, not _flake
        assert flake


class ClusterTestThingy(Test):
    """Fake ducktape test class"""

    @cluster(num_nodes=10)
    def test_bad_num_nodes(self):
        pass

    @cluster(num_nodes=0)
    def test_good_num_nodes(self):
        pass



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections

from ducktape.cluster.cluster import ClusterNode
from ducktape.cluster.cluster_spec import NodeSpec, ClusterSpec, LINUX, WINDOWS
from tests.ducktape_mock import FakeCluster

FakeRemoteAccount = collections.namedtuple('FakeRemoteAccount', ['operating_system'])


class CheckCluster(object):

    def setup_method(self, _):
        self.cluster = FakeCluster(0)
        self.cluster._available_nodes.add_node(ClusterNode(FakeRemoteAccount(operating_system=LINUX)))
        self.cluster._available_nodes.add_node(ClusterNode(FakeRemoteAccount(operating_system=LINUX)))
        self.cluster._available_nodes.add_node(ClusterNode(FakeRemoteAccount(operating_system=WINDOWS)))
        self.cluster._available_nodes.add_node(ClusterNode(FakeRemoteAccount(operating_system=WINDOWS)))
        self.cluster._available_nodes.add_node(ClusterNode(FakeRemoteAccount(operating_system=WINDOWS)))

    def spec(self, linux_nodes, windows_nodes):
        nodes = []
        for i in range(linux_nodes):
            nodes.append(NodeSpec(LINUX))
        for i in range(windows_nodes):
            nodes.append(NodeSpec(WINDOWS))
        return ClusterSpec(nodes)

    def check_enough_capacity(self):
        assert self.cluster.available().nodes.can_remove_spec(self.spec(2, 2))
        assert self.cluster.available().nodes.can_remove_spec(self.spec(2, 3))

    def check_not_enough_capacity(self):
        assert not self.cluster.available().nodes.can_remove_spec(self.spec(5, 2))
        assert not self.cluster.available().nodes.can_remove_spec(self.spec(5, 5))
        assert not self.cluster.available().nodes.can_remove_spec(self.spec(3, 3))



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


class TestStatus(object):
    def __init__(self, status):
        self._status = str(status).lower()

    def __eq__(self, other):
        return str(self).lower() == str(other).lower()

    def __str__(self):
        return self._status

    def to_json(self):
        return str(self).upper()


PASS = TestStatus("pass")
FLAKY = TestStatus("flaky")
FAIL = TestStatus("fail")
IGNORE = TestStatus("ignore")



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy

from ducktape.mark._mark import Mark

CLUSTER_SPEC_KEYWORD = "cluster_spec"
CLUSTER_SIZE_KEYWORD = "num_nodes"


class ClusterUseMetadata(Mark):
    """Provide a hint about how a given test will use the cluster."""

    def __init__(self, **kwargs):
        # shallow copy
        self.metadata = copy.copy(kwargs)

    @property
    def name(self):
        return "RESOURCE_HINT_CLUSTER_USE"

    def apply(self, seed_context, context_list):
        assert len(context_list) > 0, "cluster use annotation is not being applied to any test cases"

        for ctx in context_list:
            if not ctx.cluster_use_metadata:
                # only update if non-None and non-empty
                ctx.cluster_use_metadata = self.metadata
        return context_list


def cluster(**kwargs):
    """Test method decorator used to provide hints about how the test will use the given cluster.

    :Keywords used by ducktape:

        - ``num_nodes`` provide hint about how many nodes the test will consume
        - ``cluster_spec`` provide hint about how many nodes of each type the test will consume


    Example::

        # basic usage with num_nodes
        @cluster(num_nodes=10)
        def the_test(...):
            ...

        # basic usage with cluster_spec
        @cluster(cluster_spec=ClusterSpec.simple_linux(10))
        def the_test(...):
            ...

        # parametrized test:
        # both test cases will be marked with cluster_size of 200
        @cluster(num_nodes=200)
        @parametrize(x=1)
        @parametrize(x=2)
        def the_test(x):
            ...

        # test case {'x': 1} has cluster size 100, test case {'x': 2} has cluster size 200
        @cluster(num_nodes=100)
        @parametrize(x=1)
        @cluster(num_nodes=200)
        @parametrize(x=2)
        def the_test(x):
            ...

    """

    def cluster_use_metadata_adder(f):
        Mark.mark(f, ClusterUseMetadata(**kwargs))
        return f

    return cluster_use_metadata_adder



# Copyright 2016 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from ._mark import parametrized, Parametrize, _is_parametrize_mark
from ducktape.tests.test import TestContext


class MarkedFunctionExpander(object):
    """This class helps expand decorated/marked functions into a list of test context objects. """

    def __init__(self, session_context=None, module=None, cls=None, function=None, file=None, cluster=None):
        self.seed_context = TestContext(
            session_context=session_context, module=module, cls=cls, function=function, file=file, cluster=cluster)

        if parametrized(function):
            self.context_list = []
        else:
            self.context_list = [self.seed_context]

    def expand(self, test_parameters=None):
        """Inspect self.function for marks, and expand into a list of test context objects useable by the test runner.
        """
        f = self.seed_context.function

        # If the user has specified that they want to run tests with specific parameters, apply the parameters first,
        # then subsequently strip any parametrization decorators. Otherwise, everything gets applied normally.
        if test_parameters is not None:
            self.context_list = Parametrize(**test_parameters).apply(self.seed_context, self.context_list)

        for m in getattr(f, "marks", []):
            if test_parameters is None or not _is_parametrize_mark(m):
                self.context_list = m.apply(self.seed_context, self.context_list)

        return self.context_list



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging


class LoggerMaker(object):
    """This class helps ensure programmatically configured loggers are configured only once."""

    def __init__(self, logger_name):
        self.logger_name = logger_name

    @property
    def logger(self):
        """Read-only logger attribute."""
        if not hasattr(self, '_logger'):
            self._logger = logging.getLogger(self.logger_name)

        if not self.configured:
            self.configure_logger()

        return self._logger

    @property
    def configured(self):
        """Return True iff the logger has been configured.

        Since logging objects are global in the sense that logging.getLogger(self.logger_name) yields the same
        object, we assume it is already configured if it already has at least 1 handler.
        """
        return len(logging.getLogger(self.logger_name).handlers) > 0

    def configure_logger(self):
        raise NotImplementedError("configure_logger property must be implemented by a subclass")


def close_logger(logger):
    """Filehandles etc are not closed automatically, so close them here"""
    if logger is not None:
        handlers = logger.handlers[:]
        for handler in handlers:
            handler.close()
            logger.removeHandler(handler)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from ducktape.services.service import Service

import threading
import traceback

from six import itervalues


class BackgroundThreadService(Service):

    def __init__(self, context, num_nodes=None, cluster_spec=None, *args, **kwargs):
        super(BackgroundThreadService, self).__init__(context, num_nodes, cluster_spec, *args, **kwargs)
        self.worker_threads = {}
        self.worker_errors = {}
        self.errors = ''
        self.lock = threading.RLock()

    def _protected_worker(self, idx, node):
        """Protected worker captures exceptions and makes them available to the main thread.

        This gives us the ability to propagate exceptions thrown in background threads, if desired.
        """
        try:
            self._worker(idx, node)
        except BaseException:
            with self.lock:
                self.logger.info("BackgroundThreadService threw exception: ")
                tb = traceback.format_exc()
                self.logger.info(tb)
                self.worker_errors[threading.currentThread().name] = tb
                if self.errors:
                    self.errors += "\n"
                self.errors += "%s: %s" % (threading.currentThread().name, tb)

            raise

    def start_node(self, node):
        idx = self.idx(node)

        if idx in self.worker_threads and self.worker_threads[idx].is_alive():
            raise RuntimeError("Cannot restart node since previous thread is still alive")

        self.logger.info("Running %s node %d on %s", self.service_id, idx, node.account.hostname)
        worker = threading.Thread(
            name=self.service_id + "-worker-" + str(idx),
            target=self._protected_worker,
            args=(idx, node)
        )
        worker.daemon = True
        worker.start()
        self.worker_threads[idx] = worker

    def wait(self, timeout_sec=600):
        """Wait no more than timeout_sec for all worker threads to finish.

        raise TimeoutException if all worker threads do not finish within timeout_sec
        """
        super(BackgroundThreadService, self).wait(timeout_sec)

        self._propagate_exceptions()

    def stop(self):
        alive_workers = [worker for worker in itervalues(self.worker_threads) if worker.is_alive()]
        if len(alive_workers) > 0:
            self.logger.debug(
                "Called stop with at least one worker thread is still running: " + str(alive_workers))

            self.logger.debug("%s" % str(self.worker_threads))

        super(BackgroundThreadService, self).stop()

        self._propagate_exceptions()

    def wait_node(self, node, timeout_sec=600):
        idx = self.idx(node)
        worker_thread = self.worker_threads[idx]
        worker_thread.join(timeout_sec)
        return not(worker_thread.is_alive())

    def _propagate_exceptions(self):
        """
        Propagate exceptions thrown in background threads
        """
        with self.lock:
            if len(self.worker_errors) > 0:
                raise Exception(self.errors)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import glob
import json
import sys
from operator import attrgetter

import collections
import importlib
import inspect
import itertools
import os
import re
from typing import List

import requests
import yaml

from ducktape.tests.test import Test, TestContext
from ducktape.mark import parametrized
from ducktape.mark.mark_expander import MarkedFunctionExpander
from six import itervalues


class LoaderException(Exception):
    pass


# A helper container class
ModuleAndFile = collections.namedtuple('ModuleAndFile', ['module', 'file'])


DEFAULT_TEST_FILE_PATTERN = r"(^test_.*\.py$)|(^.*_test\.py$)"
DEFAULT_TEST_FUNCTION_PATTERN = "(^test.*)|(.*test$)"

# Included for unit tests to be able to add support for loading local file:/// URLs.
_requests_session = requests.session()


class TestLoader(object):
    """Class used to discover and load tests."""

    def __init__(self, session_context, logger, repeat=1, injected_args=None, cluster=None, subset=0, subsets=1,
                 historical_report=None):
        self.session_context = session_context
        self.cluster = cluster
        assert logger is not None
        self.logger = logger

        assert repeat >= 1
        self.repeat = repeat

        if subset >= subsets:
            raise ValueError("The subset to execute must be in the range [0, subsets-1]")
        self.subset = subset
        self.subsets = subsets

        self.historical_report = historical_report

        self.test_file_pattern = DEFAULT_TEST_FILE_PATTERN
        self.test_function_pattern = DEFAULT_TEST_FUNCTION_PATTERN

        # A non-None value here means the loader will override the injected_args
        # in any discovered test, whether or not it is parametrized
        self.injected_args = injected_args

    def load(self, symbols, excluded_test_symbols=None):
        """
        Discover tests specified by the symbols parameter (iterable of test symbols and/or test suite file paths).
        Skip any tests specified by excluded_test_symbols (iterable of test symbols).

        *Test symbol* is a pointer to the test or a group of tests.
        It is specified by the file/folder path or glob, optionally with Class.method after `::` :
        - `test-dir/`  - loads all tests under `test-dir`  but does NOT load test suites found under `test-dir`
        - `test-dir/prefix_*.py` - loads all files with a specified prefix
        - `/path/to/test/file.py`
        - `test/file.py::TestClass`
        - `test/file.py::TestClass.test_method`

        *Test suite* is a yaml file with the following format:
        ```
            # multiple test suites can be included:
            test_suite_name:
                # list included test symbols
                - path/to/test.py
            # optionally test suite can have included and excluded sections:
            # you may also specify a list of other suite's you wish to import
            # that will also be loaded when loading this file by using the
            # import tag.
            import:
                # list of yaml files whose suites will also run:
                - path/to/suite.yml
            another_test_suite:
                included:
                    # list of included test symbols:
                    - path/to/test-dir/prefix_*.py
                excluded:
                    # list of excluded test symbols:
                    - path/to/test-dir/prefix_excluded.py
        ```
        Each file found after parsing a symbol is checked to see if it contains a test:
        - Discover modules that 'look like' a test. By default, this means the filename is "test_*" or "*_test.py"
        - Discover test classes within each test module. A test class is a subclass of Test which is a leaf
          (i.e. it has no subclasses).
        - Discover test methods within each test class. A test method is a method containing 'test' in its name

        :param symbols: iterable that contains test symbols and/or test suite file paths.
        :param excluded_test_symbols: iterable that contains test symbols only.
        :return list of test context objects found during discovery. Note: if self.repeat is set to n, each test_context
            will appear in the list n times.
        """

        test_symbols = []
        test_suites = []
        # symbol can point to a test or a test suite
        for symbol in symbols:
            if symbol.endswith('.yml'):
                # if it ends with .yml, its a test suite, read included and excluded paths from the file
                test_suites.append(symbol)
            else:
                # otherwise add it to default suite's included list
                test_symbols.append(symbol)

        contexts_from_suites = self._load_test_suite_files(test_suites)
        contexts_from_symbols = self._load_test_contexts(test_symbols)
        all_included = contexts_from_suites.union(contexts_from_symbols)

        # excluded_test_symbols apply to both tests from suites and tests from symbols
        global_excluded = self._load_test_contexts(excluded_test_symbols)
        all_test_context_list = self._filter_excluded_test_contexts(all_included, global_excluded)

        # make sure no test is loaded twice
        all_test_context_list = self._filter_by_unique_test_id(all_test_context_list)

        # Sort to make sure we get a consistent order for when we create subsets
        all_test_context_list = sorted(all_test_context_list, key=attrgetter("test_id"))
        if not all_test_context_list:
            raise LoaderException("No tests to run!")
        self.logger.debug("Discovered these tests: " + str(all_test_context_list))
        # Select the subset of tests.
        if self.historical_report:
            # With timing info, try to pack the subsets reasonably evenly based on timing. To do so, get timing info
            # for each test (using avg as a fallback for missing data), sort in descending order, then start greedily
            # packing tests into bins based on the least full bin at the time.
            raw_results = _requests_session.get(self.historical_report).json()["results"]
            time_results = {r['test_id']: r['run_time_seconds'] for r in raw_results}
            avg_result_time = sum(itervalues(time_results)) / len(time_results)
            time_results = {tc.test_id: time_results.get(tc.test_id, avg_result_time) for tc in all_test_context_list}
            all_test_context_list = sorted(all_test_context_list, key=lambda x: time_results[x.test_id], reverse=True)

            subsets = [[] for _ in range(self.subsets)]
            subsets_accumulated_time = [0] * self.subsets

            for tc in all_test_context_list:
                min_subset_idx = min(range(len(subsets_accumulated_time)), key=lambda i: subsets_accumulated_time[i])
                subsets[min_subset_idx].append(tc.test_id)
                subsets_accumulated_time[min_subset_idx] += time_results[tc.test_id]

            subset_test_context_list = subsets[self.subset]
        else:
            # Without timing info, select every nth test instead of blocks of n to avoid groups of tests that are
            # parametrizations of the same test being grouped together since that can lead to a single, parameterized,
            # long-running test causing a very imbalanced workload across different subsets. Imbalance is still
            # possible, but much less likely using this heuristic.
            subset_test_context_list = list(itertools.islice(all_test_context_list, self.subset, None, self.subsets))

        self.logger.debug("Selected this subset of tests: " + str(subset_test_context_list))
        return subset_test_context_list * self.repeat

    def discover(self, directory, module_name, cls_name, method_name, injected_args=None):
        """Discover and unpack parametrized tests tied to the given module/class/method

        :return list of test_context objects
        """
        self.logger.debug("Discovering tests at {} - {} - {} - {} - {}"
                          .format(directory, module_name, cls_name, method_name, injected_args))
        # Check validity of path
        path = os.path.join(directory, module_name)
        if not os.path.exists(path):
            raise LoaderException("Path {} does not exist".format(path))

        # Recursively search path for test modules
        module_and_file = self._import_module(path)
        if module_and_file:
            # Find all tests in discovered modules and filter out any that don't match the discovery symbol
            test_context_list = self._expand_module(module_and_file)
            if len(cls_name) > 0:
                test_context_list = filter(lambda t: t.cls_name == cls_name, test_context_list)
            if len(method_name) > 0:
                test_context_list = filter(lambda t: t.function_name == method_name, test_context_list)
            if injected_args is not None:
                if isinstance(injected_args, List):
                    def condition(t):
                        return t.injected_args in injected_args
                else:
                    def condition(t):
                        return t.injected_args == injected_args
                test_context_list = filter(condition, test_context_list)

            listed = list(test_context_list)
            if not listed:
                self.logger.warn("No tests loaded for {} - {} - {} - {} - {}"
                                 .format(directory, module_name, cls_name, method_name, injected_args))
            return listed
        else:
            return []

    def _parse_discovery_symbol(self, discovery_symbol, base_dir=None):
        """Parse a single 'discovery symbol'

        :param discovery_symbol: a symbol used to target test(s).
            Looks like: <path/to/file_or_directory>[::<ClassName>[.method_name]]
        :return tuple of form (directory, module.py, cls_name, function_name)

        :raise LoaderException if it can't be parsed

        Examples:
            "path/to/directory" -> ("path/to/directory", "", "")
            "path/to/test_file.py" -> ("path/to/test_file.py", "", "")
            "path/to/test_file.py::ClassName.method" -> ("path/to/test_file.py", "ClassName", "method")
        """
        def divide_by_symbol(ds, symbol):
            if symbol not in ds:
                return ds, ""
            return ds.split(symbol, maxsplit=1)

        self.logger.debug('Trying to parse discovery symbol {}'.format(discovery_symbol))
        if base_dir:
            discovery_symbol = os.path.join(base_dir, discovery_symbol)
        if discovery_symbol.find("::") >= 0:
            path, cls_name = divide_by_symbol(discovery_symbol, "::")
            # If the part after :: contains a dot, use it to split into class + method
            cls_name, method_name = divide_by_symbol(cls_name, ".")
            method_name, injected_args_str = divide_by_symbol(method_name, "@")

            if injected_args_str:
                if self.injected_args:
                    raise LoaderException("Cannot use both global and per-method test parameters")
                try:
                    injected_args = json.loads(injected_args_str)
                except Exception as e:
                    raise LoaderException("Invalid discovery symbol: cannot parse params: " + injected_args_str) from e
            else:
                injected_args = None
        else:
            # No "::" present in symbol
            path, cls_name, method_name, injected_args = discovery_symbol, "", "", None

        return path, cls_name, method_name, injected_args

    def _import_module(self, file_path):
        """Attempt to import a python module from the file path.
        Assume file_path is an absolute path ending in '.py'

        Return the imported module..

        :param file_path: file to import module from.
        :return ModuleAndFile object that contains the successfully imported module and
            the file from which it was imported
        """
        self.logger.debug("Trying to import module at path {}".format(file_path))
        if file_path[-3:] != ".py" or not os.path.isabs(file_path):
            raise Exception("Expected absolute path ending in '.py' but got " + file_path)

        # Try all possible module imports for given file
        # Strip off '.py' before splitting
        path_pieces = [piece for piece in file_path[:-3].split("/") if len(piece) > 0]
        while len(path_pieces) > 0:
            module_name = '.'.join(path_pieces)
            # Try to import the current file as a module
            self.logger.debug("Trying to import module {}".format(module_name))
            try:
                module_and_file = ModuleAndFile(module=importlib.import_module(module_name), file=file_path)
                self.logger.debug("Successfully imported " + module_name)
                return module_and_file
            except Exception as e:
                # Because of the way we are searching for
                # valid modules in this loop, we expect some of the
                # module names we construct to fail to import.
                #
                # Therefore we check if the failure "looks normal", and log
                # expected failures only at debug level.
                #
                # Unexpected errors are aggressively logged, e.g. if the module
                # is valid but itself triggers an ImportError (e.g. typo in an
                # import line), or a SyntaxError.
                expected_error = False
                if isinstance(e, ImportError):
                    match = re.search(r"No module named '?([^\s\']+)'?", str(e))

                    if match is not None:
                        missing_module = match.groups()[0]

                        if missing_module in module_name:
                            expected_error = True
                        else:
                            # The error is still an expected error if missing_module is a suffix of module_name.
                            # This is because the error message may contain only a suffix
                            # of the original module_name if leftmost chunk of module_name is a legitimate
                            # module name, but the rightmost part doesn't exist.
                            #
                            # Check this by seeing if it is a "piecewise suffix" of module_name - i.e. if the parts
                            # delimited by dots match. This is a little bit stricter than just checking for a suffix
                            #
                            # E.g. "fancy.cool_module" is a piecewise suffix of "my.fancy.cool_module",
                            # but  "module" is not a piecewise suffix of "my.fancy.cool_module"
                            missing_module_pieces = missing_module.split(".")
                            expected_error = (missing_module_pieces == path_pieces[-len(missing_module_pieces):])

                if expected_error:
                    self.logger.debug(
                        "Failed to import %s. This is likely an artifact of the "
                        "ducktape module loading process: %s: %s", module_name, e.__class__.__name__, e)
                else:
                    self.logger.error(
                        "Failed to import %s, which may indicate a "
                        "broken test that cannot be loaded: %s: %s", module_name, e.__class__.__name__, e)
            finally:
                path_pieces = path_pieces[1:]

        self.logger.debug("Unable to import %s" % file_path)
        return None

    def _expand_module(self, module_and_file):
        """Return a list of TestContext objects, one object for every 'testable unit' in module"""

        test_context_list = []
        module = module_and_file.module
        file_name = module_and_file.file
        module_objects = module.__dict__.values()
        test_classes = [c for c in module_objects if self._is_test_class(c)]

        for cls in test_classes:
            test_context_list.extend(self._expand_class(
                TestContext(
                    session_context=self.session_context,
                    cluster=self.cluster,
                    module=module.__name__,
                    cls=cls,
                    file=file_name)))

        return test_context_list

    def _expand_class(self, t_ctx):
        """Return a list of TestContext objects, one object for each method in t_ctx.cls"""
        test_methods = []
        for f_name in dir(t_ctx.cls):
            f = getattr(t_ctx.cls, f_name)
            if self._is_test_function(f):
                test_methods.append(f)

        test_context_list = []
        for f in test_methods:
            t = t_ctx.copy(function=f)
            test_context_list.extend(self._expand_function(t))
        return test_context_list

    def _expand_function(self, t_ctx):
        expander = MarkedFunctionExpander(
            t_ctx.session_context,
            t_ctx.module,
            t_ctx.cls,
            t_ctx.function,
            t_ctx.file,
            t_ctx.cluster)
        return expander.expand(self.injected_args)

    def _find_test_files(self, path_or_glob):
        """
        Return a list of files at the specified path (or glob) that look like test files.

        - Globs are not recursive, so ** is not supported.
        - However, if the glob matches a folder (or is not a glob but simply a folder path),
            we will load all tests in that folder and recursively search the sub folders.

        :param path_or_glob: path to a test file, folder with test files or a glob that expands to folders and files
        :return: list of absolute paths to test files
        """
        test_files = []
        self.logger.debug('Looking for test files in {}'.format(path_or_glob))
        # glob is safe to be called on non-glob path - it would just return that same path wrapped in a list
        expanded_glob = glob.glob(path_or_glob)
        self.logger.debug('Expanded {} into {}'.format(path_or_glob, expanded_glob))

        def maybe_add_test_file(f):
            if self._is_test_file(f):
                test_files.append(f)
            else:
                self.logger.debug("Skipping {} because it isn't a test file".format(f))

        for path in expanded_glob:
            if not os.path.exists(path):
                raise LoaderException('Path {} does not exist'.format(path))
            self.logger.debug('Checking {}'.format(path))
            if os.path.isfile(path):
                maybe_add_test_file(path)
            elif os.path.isdir(path):
                for pwd, dirs, files in os.walk(path):
                    if "__init__.py" not in files:
                        # Not a package - ignore this directory
                        continue
                    for f in files:
                        file_path = os.path.abspath(os.path.join(pwd, f))
                        maybe_add_test_file(file_path)
            else:
                raise LoaderException("Got a path that we don't understand: " + path)

        return test_files

    def _is_test_file(self, file_name):
        """By default, a test file looks like test_*.py or *_test.py"""
        return re.match(self.test_file_pattern, os.path.basename(file_name)) is not None

    def _is_test_class(self, obj):
        """An object is a test class if it's a leafy subclass of Test."""
        return inspect.isclass(obj) and issubclass(obj, Test) and len(obj.__subclasses__()) == 0

    def _is_test_function(self, function):
        """A test function looks like a test and is callable (or expandable)."""
        if function is None:
            return False

        if not parametrized(function) and not callable(function):
            return False

        return re.match(self.test_function_pattern, function.__name__) is not None

    def _load_test_suite_files(self, test_suite_files):
        suites = list()

        suites.extend(self._read_test_suite_from_file(test_suite_files))

        all_contexts = set()
        for suite in suites:
            all_contexts.update(self._load_test_suite(**suite))
        return all_contexts

    def _load_file(self, suite_file_path):
        if not os.path.exists(suite_file_path):
            raise LoaderException(f'Path {suite_file_path} does not exist')
        if not os.path.isfile(suite_file_path):
            raise LoaderException(f'{suite_file_path} is not a file, so it cannot be a test suite')

        with open(suite_file_path) as fp:
            try:
                file_content = yaml.load(fp, Loader=yaml.FullLoader)
            except Exception as e:
                raise LoaderException("Failed to load test suite from file: " + suite_file_path, e)

        if not file_content:
            raise LoaderException("Test suite file is empty: " + suite_file_path)
        if not isinstance(file_content, dict):
            raise LoaderException("Malformed test suite file: " + suite_file_path)

        for suite_name, suite_content in file_content.items():
            if not suite_content:
                raise LoaderException("Empty test suite " + suite_name + " in " + suite_file_path)
        return file_content

    def _load_suites(self, file_path, file_content):
        suites = []
        for suite_name, suite_content in file_content.items():
            if not suite_content:
                raise LoaderException(f"Empty test suite {suite_name} in {file_path}")

            # if test suite is just a list of paths, those are included paths
            # otherwise, expect separate sections for included and excluded
            if isinstance(suite_content, list):
                included_paths = suite_content
                excluded_paths = None
            elif isinstance(suite_content, dict):
                included_paths = suite_content.get('included')
                excluded_paths = suite_content.get('excluded')
            else:
                raise LoaderException(f"Malformed test suite {suite_name} in {file_path}")
            suites.append({
                'name': suite_name,
                'included': included_paths,
                'excluded': excluded_paths,
                'base_dir': os.path.dirname(file_path)
            })
        return suites

    def _read_test_suite_from_file(self, root_suite_file_paths):
        root_suite_file_paths = [os.path.abspath(file_path) for file_path in root_suite_file_paths]
        files = {file: self._load_file(file) for file in root_suite_file_paths}
        stack = root_suite_file_paths

        # load all files
        while len(stack) != 0:
            curr = stack.pop()
            loaded = files[curr]
            if 'import' in loaded:
                if isinstance(loaded['import'], str):
                    loaded['import'] = [loaded['import']]
                directory = os.path.dirname(curr)
                # apply path of current file to the files inside
                abs_file_iter = (os.path.abspath(os.path.join(directory, file))
                                 for file in loaded.get('import', []))
                imported = [file for file in abs_file_iter if file not in files]
                for file in imported:
                    files[file] = self._load_file(file)
                stack.extend(imported)
                del files[curr]['import']

        # load all suites from all loaded files
        suites = []
        for file_name, file_context in files.items():
            suites.extend(self._load_suites(file_name, file_context))

        return suites

    def _load_test_suite(self, **kwargs):
        name = kwargs['name']
        included = kwargs['included']
        excluded = kwargs.get('excluded')
        base_dir = kwargs.get('base_dir')
        excluded_contexts = self._load_test_contexts(excluded, base_dir=base_dir)
        included_contexts = self._load_test_contexts(included, base_dir=base_dir)

        self.logger.debug("Including tests: " + str(included_contexts))
        self.logger.debug("Excluding tests: " + str(excluded_contexts))

        # filter out any excluded test from the included tests set
        all_test_context_list = self._filter_excluded_test_contexts(included_contexts, excluded_contexts)
        if not all_test_context_list:
            raise LoaderException("No tests found in  " + name)

        return all_test_context_list

    def _load_test_contexts(self, test_discovery_symbols, base_dir=None):
        """
        Load all test_context objects found in test_discovery_symbols.
        Each test discovery symbol is a dir or file path, optionally with with a ::Class or ::Class.method specified.

        :param test_discovery_symbols: list of test symbols to look into
        :return: List of test_context objects discovered by checking test_discovery_symbols (may be empty if none were
            discovered)
        """
        if not test_discovery_symbols:
            return set()
        if not isinstance(test_discovery_symbols, list):
            raise LoaderException("Expected test_discovery_symbols to be a list.")
        all_test_context_list = set()
        for symbol in test_discovery_symbols:
            path_or_glob, cls_name, method, injected_args = self._parse_discovery_symbol(symbol, base_dir)
            self.logger.debug('Parsed symbol into {} - {} - {} - {}'
                              .format(path_or_glob, cls_name, method, injected_args))
            path_or_glob = os.path.abspath(path_or_glob)

            # TODO: consider adding a check to ensure glob or dir is not used together with cls_name and method
            test_files = []
            if os.path.isfile(path_or_glob):
                # if it is a single file, just add it directly - https://github.com/confluentinc/ducktape/issues/284
                test_files = [path_or_glob]
            else:
                # otherwise, when dealing with a dir or a glob, apply pattern matching rules
                test_files = self._find_test_files(path_or_glob)

            self._add_top_level_dirs_to_sys_path(test_files)

            for test_file in test_files:
                directory = os.path.dirname(test_file)
                module_name = os.path.basename(test_file)
                test_context_list_for_file = self.discover(
                    directory, module_name, cls_name, method, injected_args=injected_args)
                all_test_context_list.update(test_context_list_for_file)
                if len(test_context_list_for_file) == 0:
                    self.logger.warn("Didn't find any tests in %s " % test_file)

        return all_test_context_list

    def _filter_by_unique_test_id(self, contexts):
        contexts_dict = dict()
        for context in contexts:
            if context.test_id not in contexts_dict:
                contexts_dict[context.test_id] = context
        return contexts_dict.values()

    def _filter_excluded_test_contexts(self, included_contexts, excluded_contexts):
        excluded_test_ids = set(map(lambda ctx: ctx.test_id, excluded_contexts))
        return set(filter(lambda ctx: ctx.test_id not in excluded_test_ids, included_contexts))

    def _add_top_level_dirs_to_sys_path(self, test_files):
        seen_dirs = set()
        for path in test_files:
            dir = os.path.dirname(path)
            while os.path.exists(os.path.join(dir, '__init__.py')):
                dir = os.path.dirname(dir)
            if dir not in seen_dirs:
                sys.path.append(dir)
                seen_dirs.add(dir)



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import pytest

from ducktape.errors import TimeoutError
from ducktape.utils.util import wait_until
import time


class CheckUtils(object):

    def check_wait_until(self):
        """Check normal wait until behavior"""
        start = time.time()

        wait_until(lambda: time.time() > start + .5, timeout_sec=2, backoff_sec=.1)

    def check_wait_until_timeout(self):
        """Check that timeout throws exception"""
        start = time.time()

        with pytest.raises(TimeoutError, match="Hello world"):
            wait_until(lambda: time.time() > start + 5, timeout_sec=.5, backoff_sec=.1, err_msg="Hello world")

    def check_wait_until_timeout_callable_msg(self):
        """Check that timeout throws exception and the error message is generated via a callable"""
        start = time.time()

        with pytest.raises(TimeoutError, match="Hello world"):
            wait_until(lambda: time.time() > start + 5, timeout_sec=.5, backoff_sec=.1, err_msg=lambda: "Hello world")

    def check_wait_until_with_exception(self):
        def condition_that_raises():
            raise Exception("OG")
        with pytest.raises(TimeoutError) as exc_info:
            wait_until(condition_that_raises, timeout_sec=.5, backoff_sec=.1, err_msg="Hello world",
                       retry_on_exc=True)
        exc_chain = exc_info.getrepr(chain=True).chain
        # 2 exceptions in the chain - OG and Hello world
        assert len(exc_chain) == 2

        # each element of a chain is a tuple of traceback, "crash" (which is the short message)
        # and optionally descr, which is None for the bottom of the chain
        # and "The exception above is ..." for all the others
        # We're interested in crash, since that one is a one-liner with the actual exception message.
        og_message = str(exc_chain[0][1])
        hello_message = str(exc_chain[1][1])
        assert "OG" in og_message
        assert "Hello world" in hello_message

    def check_wait_until_with_exception_on_first_step_only_but_still_fails(self):
        start = time.time()

        def condition_that_raises_before_3():
            if time.time() < start + .3:
                raise Exception("OG")
            else:
                return False
        with pytest.raises(TimeoutError) as exc_info:
            wait_until(condition_that_raises_before_3, timeout_sec=.5, backoff_sec=.1, err_msg="Hello world",
                       retry_on_exc=True)
        exc_chain = exc_info.getrepr(chain=True).chain
        assert len(exc_chain) == 1
        hello_message = str(exc_chain[0][1])
        assert "Hello world" in hello_message

    def check_wait_until_exception_which_succeeds_eventually(self):
        start = time.time()

        def condition_that_raises_before_3_but_then_succeeds():
            if time.time() < start + .3:
                raise Exception("OG")
            else:
                return True
        wait_until(condition_that_raises_before_3_but_then_succeeds,
                   timeout_sec=.5, backoff_sec=.1, err_msg="Hello world", retry_on_exc=True)

    def check_wait_until_breaks_early_on_exception(self):
        def condition_that_raises():
            raise Exception("OG")
        with pytest.raises(Exception, match="OG") as exc_info:
            wait_until(condition_that_raises, timeout_sec=.5, backoff_sec=.1, err_msg="Hello world")
        assert "Hello world" not in str(exc_info)



import time

from ducktape.mark.resource import cluster
from ducktape.tests.test import Test


class VariousNumNodesTest(Test):
    """
    Allocates various number of nodes.
    """

    @cluster(num_nodes=5)
    def test_five_nodes_a(self):
        assert True

    @cluster(num_nodes=5)
    def test_five_nodes_b(self):
        assert True

    @cluster(num_nodes=4)
    def test_four_nodes(self):
        assert True

    @cluster(num_nodes=3)
    def test_three_nodes_asleep(self):
        time.sleep(3)
        assert True

    @cluster(num_nodes=3)
    def test_three_nodes_a(self):
        assert True

    @cluster(num_nodes=3)
    def test_three_nodes_b(self):
        assert True

    @cluster(num_nodes=2)
    def test_two_nodes_a(self):
        assert True

    @cluster(num_nodes=2)
    def test_two_nodes_b(self):
        assert True

    @cluster(num_nodes=1)
    def test_one_node_a(self):
        assert True

    @cluster(num_nodes=1)
    def test_one_node_b(self):
        assert True



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import importlib
import json
import os
import random
from six import iteritems
import sys
import traceback

from ducktape.command_line.defaults import ConsoleDefaults
from ducktape.command_line.parse_args import parse_args
from ducktape.tests.loader import TestLoader, LoaderException
from ducktape.tests.loggermaker import close_logger
from ducktape.tests.reporter import SimpleStdoutSummaryReporter, SimpleFileSummaryReporter, \
    HTMLSummaryReporter, JSONReporter, JUnitReporter, FailedTestSymbolReporter
from ducktape.tests.runner import TestRunner
from ducktape.tests.session import SessionContext, SessionLoggerMaker
from ducktape.tests.session import generate_session_id, generate_results_dir
from ducktape.utils.local_filesystem_utils import mkdir_p
from ducktape.utils import persistence
from ducktape.utils.util import load_function


def get_user_defined_globals(globals_str):
    """Parse user-defined globals into an immutable dict using globals_str

    :param globals_str Either a file, in which case, attempt to open the file and parse the contents as JSON,
        or a JSON string representing a JSON object. The parsed JSON must represent a collection of key-value pairs,
        i.e. a python dict.
    :return dict containing user-defined global variables
    """
    if globals_str is None:
        return persistence.make_dict()

    from_file = False
    if os.path.isfile(globals_str):
        # The string appears to be a file, so try loading JSON from file
        # This may raise an IOError if the file can't be read or a ValueError if the contents of the file
        # cannot be parsed.
        user_globals = json.loads(open(globals_str, "r").read())
        from_file = True
    else:
        try:
            # try parsing directly as json if it doesn't seem to be a file
            user_globals = json.loads(globals_str)
        except ValueError as ve:
            message = str(ve)
            message += "\nglobals parameter %s is neither valid JSON nor a valid path to a JSON file." % globals_str
            raise ValueError(message)

    # Now check that the parsed JSON is a dictionary
    if not isinstance(user_globals, dict):
        if from_file:
            message = "The JSON contained in file %s must parse to a dict. " % globals_str
        else:
            message = "JSON string referred to by globals parameter must parse to a dict. "
        message += "I.e. the contents of the JSON must be an object, not an array or primitive. "
        message += "Instead found %s, which parsed to %s" % (str(user_globals), type(user_globals))

        raise ValueError(message)

    # create the immutable dict
    return persistence.make_dict(**user_globals)


def setup_results_directory(new_results_dir):
    """Make directory in which results will be stored"""
    if os.path.exists(new_results_dir):
        raise Exception(
            "A file or directory at %s already exists. Exiting without overwriting." % new_results_dir)
    mkdir_p(new_results_dir)


def update_latest_symlink(results_root, new_results_dir):
    """Create or update symlink "latest" which points to the new test results directory"""
    latest_test_dir = os.path.join(results_root, "latest")
    if os.path.islink(latest_test_dir):
        os.unlink(latest_test_dir)
    os.symlink(new_results_dir, latest_test_dir)


def main():
    """Ducktape entry point. This contains top level logic for ducktape command-line program which does the following:

        Discover tests
        Initialize cluster for distributed services
        Run tests
        Report a summary of all results
    """
    args_dict = parse_args(sys.argv[1:])

    injected_args = None
    if args_dict["parameters"]:
        try:
            injected_args = json.loads(args_dict["parameters"])
        except ValueError as e:
            print("parameters are not valid json: " + str(e))
            sys.exit(1)

    args_dict["globals"] = get_user_defined_globals(args_dict.get("globals"))

    # Make .ducktape directory where metadata such as the last used session_id is stored
    if not os.path.isdir(ConsoleDefaults.METADATA_DIR):
        os.makedirs(ConsoleDefaults.METADATA_DIR)

    # Generate a shared 'global' identifier for this test run and create the directory
    # in which all test results will be stored
    session_id = generate_session_id(ConsoleDefaults.SESSION_ID_FILE)
    results_dir = generate_results_dir(args_dict["results_root"], session_id)
    setup_results_directory(results_dir)

    session_context = SessionContext(session_id=session_id, results_dir=results_dir, **args_dict)
    session_logger = SessionLoggerMaker(session_context).logger
    for k, v in iteritems(args_dict):
        session_logger.debug("Configuration: %s=%s", k, v)

    # Discover and load tests to be run
    loader = TestLoader(session_context, session_logger, repeat=args_dict["repeat"], injected_args=injected_args,
                        subset=args_dict["subset"], subsets=args_dict["subsets"])
    try:
        tests = loader.load(args_dict["test_path"], excluded_test_symbols=args_dict['exclude'])
    except LoaderException as e:
        print("Failed while trying to discover tests: {}".format(e))
        sys.exit(1)

    if args_dict["collect_only"]:
        print("Collected %d tests:" % len(tests))
        for test in tests:
            print("    " + str(test))
        sys.exit(0)

    if args_dict["collect_num_nodes"]:
        total_nodes = sum(test.expected_num_nodes for test in tests)
        print(total_nodes)
        sys.exit(0)

    if args_dict["sample"]:
        print("Running a sample of %d tests" % args_dict["sample"])
        try:
            tests = random.sample(tests, args_dict["sample"])
        except ValueError as e:
            if args_dict["sample"] > len(tests):
                print("sample size %d greater than number of tests %d; running all tests" % (
                    args_dict["sample"], len(tests)))
            else:
                print("invalid sample size (%s), running all tests" % e)

    # Initializing the cluster is slow, so do so only if
    # tests are sure to be run
    try:
        (cluster_mod_name, cluster_class_name) = args_dict["cluster"].rsplit('.', 1)
        cluster_mod = importlib.import_module(cluster_mod_name)
        cluster_class = getattr(cluster_mod, cluster_class_name)

        cluster_kwargs = {"cluster_file": args_dict["cluster_file"]}
        checker_function_names = args_dict['ssh_checker_function']
        if checker_function_names:
            checkers = [load_function(func_path) for func_path in checker_function_names]
            if checkers:
                cluster_kwargs['ssh_exception_checks'] = checkers
        cluster = cluster_class(**cluster_kwargs)
        for ctx in tests:
            # Note that we're attaching a reference to cluster
            # only after test context objects have been instantiated
            ctx.cluster = cluster
    except Exception:
        print("Failed to load cluster: ", str(sys.exc_info()[0]))
        print(traceback.format_exc(limit=16))
        sys.exit(1)

    # Run the tests
    deflake_num = args_dict['deflake']
    if deflake_num < 1:
        session_logger.warning("specified number of deflake runs specified to be less than 1, running without deflake.")
    deflake_num = max(1, deflake_num)
    runner = TestRunner(cluster, session_context, session_logger, tests, deflake_num)
    test_results = runner.run_all_tests()

    # Report results
    reporters = [
        SimpleStdoutSummaryReporter(test_results),
        SimpleFileSummaryReporter(test_results),
        HTMLSummaryReporter(test_results),
        JSONReporter(test_results),
        JUnitReporter(test_results),
        FailedTestSymbolReporter(test_results)
    ]

    for r in reporters:
        r.report()

    update_latest_symlink(args_dict["results_root"], results_dir)
    close_logger(session_logger)
    if not test_results.get_aggregate_success():
        # Non-zero exit if at least one test failed
        sys.exit(1)



# Copyright 2014 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import

from .json import JsonCluster, make_remote_account
import json
import os
from .remoteaccount import RemoteAccountSSHConfig
import subprocess
from ducktape.json_serializable import DucktapeJSONEncoder


class VagrantCluster(JsonCluster):
    """
    An implementation of Cluster that uses a set of VMs created by Vagrant. Because we need hostnames that can be
    advertised, this assumes that the Vagrant VM's name is a routeable hostname on all the hosts.

    - If cluster_file is specified in the constructor's kwargs (i.e. passed via command line argument --cluster-file)
      - If cluster_file exists on the filesystem, read cluster info from the file
      - Otherwise, retrieve cluster info via "vagrant ssh-config" from vagrant and write cluster info to cluster_file
    - Otherwise, retrieve cluster info via "vagrant ssh-config" from vagrant
    """

    def __init__(self, *args, make_remote_account_func=make_remote_account, **kwargs):
        is_read_from_file = False
        self.ssh_exception_checks = kwargs.get("ssh_exception_checks")
        cluster_file = kwargs.get("cluster_file")
        if cluster_file is not None:
            try:
                cluster_json = json.load(open(os.path.abspath(cluster_file)))
                is_read_from_file = True
            except IOError:
                # It is OK if file is not found. Call vagrant ssh-info to read the cluster info.
                pass

        if not is_read_from_file:
            cluster_json = {
                "nodes": self._get_nodes_from_vagrant(make_remote_account_func)
            }

        super(VagrantCluster, self).__init__(
            cluster_json, *args, make_remote_account_func=make_remote_account_func, **kwargs)

        # If cluster file is specified but the cluster info is not read from it, write the cluster info into the file
        if not is_read_from_file and cluster_file is not None:
            nodes = [
                {
                    "ssh_config": node_account.ssh_config,
                    "externally_routable_ip": node_account.externally_routable_ip
                }
                for node_account in self._available_accounts
            ]
            cluster_json["nodes"] = nodes
            with open(cluster_file, 'w+') as fd:
                json.dump(cluster_json, fd, cls=DucktapeJSONEncoder, indent=2, separators=(',', ': '), sort_keys=True)

        # Release any ssh clients used in querying the nodes for metadata
        for node_account in self._available_accounts:
            node_account.close()

    def _get_nodes_from_vagrant(self, make_remote_account_func):
        ssh_config_info, error = self._vagrant_ssh_config()

        nodes = []
        node_info_arr = ssh_config_info.split("\n\n")
        node_info_arr = [ninfo.strip() for ninfo in node_info_arr if ninfo.strip()]

        for ninfo in node_info_arr:
            ssh_config = RemoteAccountSSHConfig.from_string(ninfo)

            account = None
            try:
                account = make_remote_account_func(ssh_config, ssh_exception_checks=self.ssh_exception_checks)
                externally_routable_ip = account.fetch_externally_routable_ip()
            finally:
                if account:
                    account.close()
                    del account

            nodes.append({
                "ssh_config": ssh_config.to_json(),
                "externally_routable_ip": externally_routable_ip
            })

        return nodes

    def _vagrant_ssh_config(self):
        ssh_config_info, error = subprocess.Popen("vagrant ssh-config", shell=True, stdout=subprocess.PIPE,
                                                  stderr=subprocess.PIPE, close_fds=True,
                                                  # Force to text mode in py2/3 compatible way
                                                  universal_newlines=True).communicate()
        return ssh_config_info, error



# Copyright 2015 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
import os
import random
import shutil
import sys
import tempfile

from ducktape.cluster.cluster_spec import ClusterSpec
from ducktape.tests.test import Test, TestContext, _escape_pathname, _compress_cmd, in_dir, in_temp_dir
from tests import ducktape_mock


class DummyTest(Test):
    """class description"""

    def test_class_description(self):
        pass

    def test_function_description(self):
        """function description"""
        pass


class DummyTestNoDescription(Test):
    def test_this(self):
        pass


class CheckLifecycle(object):
    def check_test_context_double_close(self):
        context = TestContext(session_context=ducktape_mock.session_context(),
                              cls=DummyTest, function=DummyTest.test_function_description)
        context.close()
        context.close()
        assert not hasattr(context, "services")

    def check_cluster_property(self):
        exp_cluster = ClusterSpec.simple_linux(5)
        tc = TestContext(session_context=ducktape_mock.session_context(), cluster=exp_cluster,
                         cls=DummyTest, function=DummyTest.test_function_description)
        test_obj = tc.cls(tc)
        assert test_obj.cluster == exp_cluster


class CheckEscapePathname(object):

    def check_illegal_path(self):
        path = "\\/.a=2,   b=x/y/z"
        assert _escape_pathname(path) == "a=2.b=x.y.z"

    def check_negative(self):
        # it's better if negative numbers are preserved
        path = "x= -2, y=-50"
        assert _escape_pathname(path) == "x=-2.y=-50"

    def check_many_dots(self):
        path = "..a.....b.c...d."
        assert _escape_pathname(path) == "a.b.c.d"


class CheckDescription(object):
    """Check that pulling a description from a test works as expected."""

    def check_from_function(self):
        """If the function has a docstring, the description should come from the function"""
        context = TestContext(session_context=ducktape_mock.session_context(),
                              cls=DummyTest, function=DummyTest.test_function_description)
        assert context.description == "function description"

    def check_from_class(self):
        """If the test method has no docstring, description should come from the class docstring"""
        context = TestContext(session_context=ducktape_mock.session_context(),
                              cls=DummyTest, function=DummyTest.test_class_description)
        assert context.description == "class description"

    def check_no_description(self):
        """If nobody has a docstring, there shouldn't be an error, and description should be empty string"""
        context = TestContext(session_context=ducktape_mock.session_context(),
                              cls=DummyTestNoDescription, function=DummyTestNoDescription.test_this)
        assert context.description == ""


class CheckCompressCmd(object):
    """Check expected behavior of compress command used before collecting service logs"""

    def setup_method(self, _):
        self.tempdir = tempfile.mkdtemp()

    def _make_random_file(self, dir, num_chars=10000):
        """Populate filename with random characters."""
        filename = os.path.join(dir, "f-%d" % random.randint(1, 2**63 - 1))
        content = "".join([random.choice("0123456789abcdefghijklmnopqrstuvwxyz\n") for _ in range(num_chars)])
        with open(filename, "w") as f:
            f.writelines(content)
        return filename

    def _make_files(self, dir, num_files=10):
        """Populate dir with several files with random characters."""
        for i in range(num_files):
            self._make_random_file(dir)

    def _validate_compressed(self, uncompressed_path):
        if uncompressed_path.endswith(os.path.sep):
            uncompressed_path = uncompressed_path[:-len(os.path.sep)]

        compressed_path = uncompressed_path + ".tgz"

        # verify original file is replaced by filename.tgz
        assert os.path.exists(compressed_path)
        assert not os.path.exists(uncompressed_path)

        # verify that uncompressing gets us back the original
        with in_dir(self.tempdir):
            os.system("tar xzf %s" % (compressed_path))
            assert os.path.exists(uncompressed_path)

    def check_compress_service_logs_swallow_error(self):
        """Try compressing a non-existent service log, and check that it logs a message without throwing an error.
        """
        from tests.ducktape_mock import session_context
        tc = TestContext(
            session_context=session_context(),
            module=sys.modules[DummyTestNoDescription.__module__],
            cls=DummyTestNoDescription,
            function=DummyTestNoDescription.test_this
        )

        tc._logger = logging.getLogger(__name__)
        temp_log_file = tempfile.NamedTemporaryFile(delete=False).name

        try:
            tmp_log_handler = logging.FileHandler(temp_log_file)
            tc._logger.addHandler(tmp_log_handler)

            test_obj = tc.cls(tc)

            # Expect an error to be triggered but swallowed
            test_obj.compress_service_logs(node=None, service=None, node_logs=["hi"])

            tmp_log_handler.close()
            with open(temp_log_file, "r") as f:
                s = f.read()
                assert s.find("Error compressing log hi") >= 0
        finally:
            if os.path.exists(temp_log_file):
                os.remove(temp_log_file)

    def check_abs_path_file(self):
        """Check compress command on an absolute path to a file"""
        with in_temp_dir() as dir1:
            filename = self._make_random_file(self.tempdir)
            abspath_filename = os.path.abspath(filename)

            # since we're using absolute path to file, we should be able to run the compress command from anywhere
            with in_temp_dir() as dir2:
                assert dir1 != dir2

                os.system(_compress_cmd(abspath_filename))
                self._validate_compressed(abspath_filename)

    def check_relative_path_file(self):
        """Check compress command on a relative path to a file"""
        with in_temp_dir():
            filename = self._make_random_file(self.tempdir)
            with in_dir(self.tempdir):
                filename = os.path.basename(filename)
                assert len(filename.split(os.path.sep)) == 1

                # compress it!
                os.system(_compress_cmd(filename))
                self._validate_compressed(filename)

    def check_abs_path_dir(self):
        """Validate compress command with absolute path to a directory"""
        dirname = tempfile.mkdtemp(dir=self.tempdir)
        dirname = os.path.abspath(dirname)
        self._make_files(dirname)

        # compress it!
        if not dirname.endswith(os.path.sep):
            # extra check - ensure all of this works with trailing '/'
            dirname += os.path.sep

        os.system(_compress_cmd(dirname))
        self._validate_compressed(dirname)

    def check_relative_path_dir(self):
        """Validate tarball compression of a directory"""
        dirname = tempfile.mkdtemp(dir=self.tempdir)
        self._make_files(dirname)

        dirname = os.path.basename(dirname)
        assert len(dirname.split(os.path.sep)) == 1

        # compress it!
        with in_dir(self.tempdir):
            os.system(_compress_cmd(dirname))
            self._validate_compressed(dirname)

    def teardown_method(self, _):
        if os.path.exists(self.tempdir):
            shutil.rmtree(self.tempdir)



# Copyright 2017 Confluent Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Dict, List, Tuple

from six import iteritems

from ducktape.cluster.remoteaccount import RemoteAccount


class NodeNotPresentError(Exception):
    pass


class InsufficientResourcesError(Exception):
    pass


class InsufficientHealthyNodesError(InsufficientResourcesError):

    def __init__(self, bad_nodes: List, *args):
        self.bad_nodes = bad_nodes
        super().__init__(*args)


class NodeContainer(object):
    os_to_nodes: Dict = None

    def __init__(self, nodes: List = None):
        """
        Create a NodeContainer with the given nodes.

        Node objects should implement at least an operating_system property.

        :param nodes:           A collection of node objects to add, or None to add nothing.
        """
        self.os_to_nodes = {}
        if nodes is not None:
            for node in nodes:
                self.os_to_nodes.setdefault(node.operating_system, []).append(node)

    def size(self):
        """
        Returns the total number of nodes in the container.
        """
        return sum([len(val) for val in self.os_to_nodes.values()])

    def __len__(self):
        return self.size()

    def __iter__(self):
        return self.elements()

    def elements(self, operating_system=None):
        """
        Yield the elements in this container.

        :param operating_system:    If this is non-None, we will iterate only over elements
                                    which have this operating system.
        """
        if operating_system is None:
            for node_list in self.os_to_nodes.values():
                for node in node_list:
                    yield node
        else:
            for node in self.os_to_nodes.get(operating_system, []):
                yield node

    def add_node(self, node):
        """
        Add a node to this collection.

        :param node:                        The node to add.
        """
        self.os_to_nodes.setdefault(node.operating_system, []).append(node)

    def add_nodes(self, nodes):
        """
        Add a collection of nodes to this collection.

        :param nodes:                       The nodes to add.
        """
        for node in nodes:
            self.add_node(node)

    def remove_node(self, node):
        """
        Removes a node from this collection.

        :param node:                        The node to remove.
        :returns:                           The node which has been removed.
        :throws NodeNotPresentError:        If the node is not in the collection.
        """
        try:
            return self.os_to_nodes.get(node.operating_system, []).remove(node)
        except ValueError:
            raise NodeNotPresentError

    def remove_nodes(self, nodes):
        """
        Remove a collection of nodes from this collection.

        :param nodes:                       The nodes to remove.
        """
        for node in nodes:
            self.remove_node(node)

    def remove_spec(self, cluster_spec) -> Tuple[List, List]:
        """
        Remove nodes matching a ClusterSpec from this NodeContainer.

        :param cluster_spec:                    The cluster spec.  This will not be modified.
        :returns:                               List of good nodes and a list of bad nodes.
        :raises:                                InsufficientResourcesError when there aren't enough total nodes
                                                InsufficientHealthyNodesError when there aren't enough healthy nodes
        """

        err = self.attempt_remove_spec(cluster_spec)
        if err:
            # there weren't enough nodes to even attempt allocations, raise the exception
            raise InsufficientResourcesError(err)

        good_nodes = []
        bad_nodes = []
        msg = ""
        # we have enough nodes for each OS, now try allocating while doing health checks if nodes support them
        for os, node_specs in iteritems(cluster_spec.nodes.os_to_nodes):
            num_nodes = len(node_specs)
            good_per_os = []
            avail_nodes = self.os_to_nodes.get(os, [])
            # loop over all available nodes
            # for i in range(0, len(avail_nodes)):
            while avail_nodes and (len(good_per_os) < num_nodes):
                node = avail_nodes.pop(0)
                if isinstance(node, RemoteAccount):
                    if node.available():
                        good_per_os.append(node)
                    else:
                        bad_nodes.append(node)
                else:
                    good_per_os.append(node)

            good_nodes.extend(good_per_os)
            # if we don't have enough good nodes to allocate for this OS,
            # set the status as failed
            if len(good_per_os) < num_nodes:
                msg += f"{os} nodes requested: {num_nodes}. Healthy {os} nodes available: {len(good_per_os)}"

        # we didn't have enough healthy nodes for at least one of the OS-s
        # no need to keep the allocated nodes, since there aren't enough of them
        # let's return good ones back to this container
        # and raise the exception with bad ones
        if msg:
            for node in good_nodes:
                self.add_node(node)
            raise InsufficientHealthyNodesError(bad_nodes, msg)

        return good_nodes, bad_nodes

    def can_remove_spec(self, cluster_spec):
        """
        Determine if we can remove nodes matching a ClusterSpec from this NodeContainer.
        This container will not be modified.

        :param cluster_spec:                    The cluster spec.  This will not be modified.
        :returns:                               True if we could remove the nodes; false otherwise
        """
        msg = self.attempt_remove_spec(cluster_spec)
        return len(msg) == 0

    def attempt_remove_spec(self, cluster_spec):
        """
        Attempt to remove a cluster_spec from this node container.

        :param cluster_spec:                    The cluster spec.  This will not be modified.
        :returns:                               An empty string if we can remove the nodes;
                                                an error string otherwise.
        """
        msg = ""
        for os, node_specs in iteritems(cluster_spec.nodes.os_to_nodes):
            num_nodes = len(node_specs)
            avail_nodes = len(self.os_to_nodes.get(os, []))
            if avail_nodes < num_nodes:
                msg = msg + "%s nodes requested: %d. %s nodes available: %d" % \
                            (os, num_nodes, os, avail_nodes)
        return msg

    def clone(self):
        """
        Returns a deep copy of this object.
        """
        container = NodeContainer()
        for operating_system, nodes in iteritems(self.os_to_nodes):
            for node in nodes:
                container.os_to_nodes.setdefault(operating_system, []).append(node)
        return container
